//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Material/MaterialEditorDialog.cpp
//! @brief     Implements class MaterialEditorDialog.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Material/MaterialEditorDialog.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Model/Sample/MaterialModel.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Support/Style/Style.h"
#include "GUI/View/Material/MaterialEditorModel.h"
#include "GUI/View/Numeric/NumberUtil.h"
#include "GUI/View/Tool/mainwindow_constants.h"
#include "GUI/View/Widget/ApplicationSettings.h"
#include "GUI/View/Widget/StyledToolbar.h"
#include <QAction>
#include <QColorDialog>
#include <QDialog>
#include <QDialogButtonBox>
#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QMessageBox>
#include <QPushButton>
#include <QSettings>
#include <QTreeView>
#include <QVBoxLayout>
#include <QWidget>

//! Dialog to select a material and also to edit the list of existing materials.
//! The dialog operates on a copy of the current materials. The original material store is only
//! updated if changes have been made and the dialog has been closed with OK.
//! Use this dialog only with the two static methods to edit the list, or to choose a material
class MaterialEditorDialog : public QDialog {
public:
    MaterialEditorDialog(SampleItem* sample, const QString& identifier);
    ~MaterialEditorDialog() override;

    MaterialItem* currentMaterialItem();

private:
    //! updates original material model with the edited model
    void accept() override;

    void addRefractiveMaterial();
    void addSldMaterial();
    void cloneCurrentMaterial();
    void removeCurrentMaterial();

    void setCurrentMaterial(const MaterialItem* m);

    void selectColor();

    void onSelectColor();
    void onSelectMaterial();

    void onChangeCurrent();
    QModelIndex currentIndex() const;

    //! Returns the list of material identifiers of the materials currently used in the sample.
    //! E.g. the material selected in a particle.
    QStringList identifiersOfUsedMaterials() const;

    MaterialModel m_tmpMaterialModel; //! Materials for working within this dialog

    QAction* m_cloneMaterialAction;
    QAction* m_removeMaterialAction;

    MaterialEditorModel* m_model; //! Model for the left list. Works on m_tmpMaterialModel
    SampleItem* m_sample;

    QTreeView* m_treeView;
    QPushButton* m_selectColorButton;
    QWidget* m_editor;
    QGroupBox* m_sld_group;
    QGroupBox* m_refr_group;
    QDoubleSpinBox* m_xSpinBox;
    QDoubleSpinBox* m_ySpinBox;
    QDoubleSpinBox* m_zSpinBox;
    QLineEdit* m_deltaEdit;
    QLineEdit* m_betaEdit;
    QLineEdit* m_realEdit;
    QLineEdit* m_imaginaryEdit;
    QLineEdit* m_nameEdit;
    QLineEdit* m_colorInfo;
};


MaterialEditorDialog::MaterialEditorDialog(SampleItem* sample, const QString& identifier)
    : m_sample(sample)
{
    setObjectName("MaterialEditorDialog");

    m_tmpMaterialModel.initFrom(m_sample->materialModel());
    m_model = new MaterialEditorModel(&m_tmpMaterialModel);

    setGeometry(0, 0, 1023, 469);
    setWindowTitle("Material Editor");

    auto* layout = new QVBoxLayout;
    setLayout(layout);

    auto* hlayout = new QHBoxLayout;
    layout->addLayout(hlayout);

    //... Left side: list of materials

    m_treeView = new QTreeView;
    hlayout->addWidget(m_treeView);
    m_treeView->setContextMenuPolicy(Qt::ActionsContextMenu);
    m_treeView->setSelectionMode(QAbstractItemView::SingleSelection);
    m_treeView->setSelectionBehavior(QAbstractItemView::SelectRows);
    m_treeView->setRootIsDecorated(false);

    //... Right side: editor for one material

    m_editor = new QWidget;
    hlayout->addWidget(m_editor);
    auto* ed_vlayout = new QVBoxLayout;
    m_editor->setLayout(ed_vlayout);
    ed_vlayout->setSpacing(6);
    ed_vlayout->setContentsMargins(0, 0, 0, 0);

    // Name and color

    auto* nc_form = new QFormLayout;
    ed_vlayout->addLayout(nc_form);
    nc_form->setWidget(0, QFormLayout::LabelRole, new QLabel("Name:"));
    m_nameEdit = new QLineEdit;
    nc_form->setWidget(0, QFormLayout::FieldRole, m_nameEdit);

    nc_form->setWidget(1, QFormLayout::LabelRole, new QLabel("Color:"));
    auto* colorLine = new QHBoxLayout;
    nc_form->setLayout(1, QFormLayout::FieldRole, colorLine);

    m_selectColorButton = new QPushButton;
    colorLine->addWidget(m_selectColorButton);
    QSizePolicy sizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
    sizePolicy.setHorizontalStretch(0);
    sizePolicy.setVerticalStretch(0);
    sizePolicy.setHeightForWidth(m_selectColorButton->sizePolicy().hasHeightForWidth());
    m_selectColorButton->setSizePolicy(sizePolicy);
    m_selectColorButton->setToolTip("Choose color");

    m_colorInfo = new QLineEdit;
    colorLine->addWidget(m_colorInfo);
    m_colorInfo->setEnabled(false);

    // Refractive data

    m_refr_group = new QGroupBox("Material data (refractive index based)");
    ed_vlayout->addWidget(m_refr_group);

    auto* refr_form = new QFormLayout;
    m_refr_group->setLayout(refr_form);
    m_deltaEdit = new QLineEdit;
    m_deltaEdit->setToolTip("Delta of refractive index (n = 1 - delta + i*beta)");
    m_betaEdit = new QLineEdit;
    m_betaEdit->setToolTip("Beta of refractive index (n = 1 - delta + i*beta)");
    refr_form->addRow("Delta:", m_deltaEdit);
    refr_form->addRow("Beta:", m_betaEdit);

    m_sld_group = new QGroupBox("Material data (SLD based [1/Å²])");
    ed_vlayout->addWidget(m_sld_group);

    auto* sld_form = new QFormLayout;
    m_sld_group->setLayout(sld_form);
    m_realEdit = new QLineEdit;
    m_realEdit->setToolTip("Real part of SLD (SLD = real - i*imag), AA^{-2}");
    m_imaginaryEdit = new QLineEdit;
    m_imaginaryEdit->setToolTip("Imaginary part of SLD (SLD = real - i*imag), AA^{-2}");
    sld_form->addRow("Real:", m_realEdit);
    sld_form->addRow("Imaginary:", m_imaginaryEdit);

    // Magnetization

    auto* groupBox_3 = new QGroupBox("Magnetization");
    ed_vlayout->addWidget(groupBox_3);
    auto* formLayout_2 = new QFormLayout(groupBox_3);

    m_xSpinBox = new QDoubleSpinBox;
    formLayout_2->addRow("X:", m_xSpinBox);
    m_xSpinBox->setSuffix(" A/m");
    m_xSpinBox->setToolTip("x coordinate of magnetization");

    m_ySpinBox = new QDoubleSpinBox;
    formLayout_2->addRow("Y:", m_ySpinBox);
    m_ySpinBox->setSuffix(" A/m");
    m_ySpinBox->setToolTip("y coordinate of magnetization");

    m_zSpinBox = new QDoubleSpinBox;
    formLayout_2->addRow("Z:", m_zSpinBox);
    m_zSpinBox->setSuffix(" A/m");
    m_zSpinBox->setToolTip("z coordinate of magnetization");

    //... Buttons at bottom

    auto* verticalSpacer = new QSpacerItem(20, 15, QSizePolicy::Minimum, QSizePolicy::Expanding);
    ed_vlayout->addItem(verticalSpacer);

    auto* m_buttonBox = new QDialogButtonBox;
    layout->addWidget(m_buttonBox);
    m_buttonBox->setOrientation(Qt::Horizontal);
    m_buttonBox->setStandardButtons(QDialogButtonBox::Cancel | QDialogButtonBox::Ok);

    //... Configure

    using GUI::View::NumberUtil::configScientificDoubleEdit;
    configScientificDoubleEdit(m_deltaEdit, RealLimits::limitless());
    configScientificDoubleEdit(m_betaEdit, RealLimits::limitless());
    configScientificDoubleEdit(m_realEdit, RealLimits::limitless());
    configScientificDoubleEdit(m_imaginaryEdit, RealLimits::limitless());

    using GUI::View::NumberUtil::configSpinBox;
    configSpinBox(m_xSpinBox, 3, RealLimits::limitless());
    configSpinBox(m_ySpinBox, 3, RealLimits::limitless());
    configSpinBox(m_zSpinBox, 3, RealLimits::limitless());

    // Setting z-component is temporary disabled (see issue #654)
    // When implemented, rm disabling
    m_zSpinBox->setDisabled(true);

    auto* addRefractiveMaterialAction = new QAction("Add material (refractive index)");
    addRefractiveMaterialAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    addRefractiveMaterialAction->setToolTip("Add new material");
    connect(addRefractiveMaterialAction, &QAction::triggered, this,
            &MaterialEditorDialog::addRefractiveMaterial);

    auto* addSldMaterialAction = new QAction("Add material (SLD)");
    addSldMaterialAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    addSldMaterialAction->setToolTip("Add new material");
    connect(addSldMaterialAction, &QAction::triggered, this, &MaterialEditorDialog::addSldMaterial);

    m_cloneMaterialAction = new QAction("Copy");
    m_cloneMaterialAction->setIcon(QIcon(":/images/content-copy.svg"));
    m_cloneMaterialAction->setToolTip("Copy selected material");
    connect(m_cloneMaterialAction, &QAction::triggered, this,
            &MaterialEditorDialog::cloneCurrentMaterial);

    m_removeMaterialAction = new QAction("Remove");
    m_removeMaterialAction->setIcon(QIcon(":/images/delete.svg"));
    m_removeMaterialAction->setToolTip("Remove selected material");
    connect(m_removeMaterialAction, &QAction::triggered, this,
            &MaterialEditorDialog::removeCurrentMaterial);

    m_treeView->addAction(addRefractiveMaterialAction);
    m_treeView->addAction(addSldMaterialAction);
    m_treeView->addAction(m_cloneMaterialAction);
    auto* separator = new QAction;
    separator->setSeparator(true);
    m_treeView->addAction(separator);
    m_treeView->addAction(m_removeMaterialAction);
    m_treeView->setModel(m_model);
    m_treeView->header()->setSectionResizeMode(QHeaderView::ResizeToContents);

    auto* toolbar = new StyledToolbar;
    toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    toolbar->addAction(addRefractiveMaterialAction);
    toolbar->addAction(addSldMaterialAction);
    toolbar->addAction(m_cloneMaterialAction);
    toolbar->addAction(m_removeMaterialAction);
    layout->insertWidget(0, toolbar);

    GUI::Style::setResizable(this);
    appSettings->loadWindowSizeAndPos(this);

    connect(m_treeView->selectionModel(), &QItemSelectionModel::currentChanged, this,
            &MaterialEditorDialog::onChangeCurrent);

    connect(m_selectColorButton, &QPushButton::clicked, this, &MaterialEditorDialog::selectColor);

    connect(m_nameEdit, &QLineEdit::textEdited,
            [&](const QString& t) { m_model->setMaterialItemName(currentIndex(), t); });

    connect(m_xSpinBox, &QDoubleSpinBox::valueChanged,
            [&](double value) { m_model->setX(currentIndex(), value); });

    connect(m_ySpinBox, &QDoubleSpinBox::valueChanged,
            [&](double value) { m_model->setY(currentIndex(), value); });

    connect(m_zSpinBox, &QDoubleSpinBox::valueChanged,
            [&](double value) { m_model->setZ(currentIndex(), value); });

    connect(m_deltaEdit, &QLineEdit::editingFinished,
            [&]() { m_model->setDelta(currentIndex(), m_deltaEdit->text().toDouble()); });

    connect(m_betaEdit, &QLineEdit::editingFinished,
            [&]() { m_model->setBeta(currentIndex(), m_betaEdit->text().toDouble()); });

    connect(m_realEdit, &QLineEdit::editingFinished,
            [&]() { m_model->setRe(currentIndex(), m_realEdit->text().toDouble()); });

    connect(m_imaginaryEdit, &QLineEdit::editingFinished,
            [&]() { m_model->setIm(currentIndex(), m_imaginaryEdit->text().toDouble()); });

    connect(m_buttonBox, &QDialogButtonBox::accepted, this, &MaterialEditorDialog::accept);
    connect(m_buttonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);

    if (m_model->rowCount() > 0)
        m_treeView->setCurrentIndex(m_model->first());
    else {
        m_sld_group->hide();
        m_editor->setEnabled(false);
    }

    m_treeView->setCurrentIndex(m_model->indexFromMaterial(identifier));
    onSelectMaterial();
}

MaterialEditorDialog::~MaterialEditorDialog()
{
    appSettings->saveWindowSizeAndPos(this);
}

void MaterialEditorDialog::accept()
{
    m_sample->materialModel().initFrom(m_tmpMaterialModel);
    QDialog::accept();
}

void MaterialEditorDialog::addRefractiveMaterial()
{
    setCurrentMaterial(m_model->addRefractiveMaterialItem("unnamed", 0.0, 0.0));
}

void MaterialEditorDialog::addSldMaterial()
{
    setCurrentMaterial(m_model->addSLDMaterialItem("unnamed", 0.0, 0.0));
}

void MaterialEditorDialog::cloneCurrentMaterial()
{
    if (currentIndex().isValid())
        setCurrentMaterial(m_model->cloneMaterialItem(currentIndex()));
}

void MaterialEditorDialog::removeCurrentMaterial()
{
    MaterialItem* material = currentMaterialItem();
    if (!material)
        return;

    if (identifiersOfUsedMaterials().contains(material->identifier())) {
        QMessageBox::warning(this, "Remove material",
                             "This material cannot be removed - it is used in the current sample.");
        return;
    }

    m_model->removeMaterial(currentIndex());
}

void MaterialEditorDialog::selectColor()
{
    if (auto newColor = QColorDialog::getColor(currentMaterialItem()->color()); newColor.isValid())
        m_model->setColor(currentIndex(), newColor);
    onSelectColor();
}

void MaterialEditorDialog::onSelectColor()
{
    auto* materialItem = currentMaterialItem();
    m_nameEdit->setText(materialItem->matItemName());
    m_colorInfo->setText(QString("[%1, %2, %3] (%4)")
                             .arg(materialItem->color().red())
                             .arg(materialItem->color().green())
                             .arg(materialItem->color().blue())
                             .arg(materialItem->color().alpha()));
    QPixmap pixmap(m_selectColorButton->iconSize());
    pixmap.fill(materialItem->color());
    m_selectColorButton->setIcon(pixmap);
}

void MaterialEditorDialog::onSelectMaterial()
{
    MaterialItem* materialItem = currentMaterialItem();
    if (!materialItem)
        return;

    m_refr_group->setVisible(materialItem->hasRefractiveIndex());
    m_sld_group->setVisible(!materialItem->hasRefractiveIndex());

    m_editor->setEnabled(materialItem != nullptr);
    if (materialItem == nullptr) {
        m_refr_group->show();
        m_sld_group->hide();
        for (auto* lineEdit : m_editor->findChildren<QLineEdit*>())
            lineEdit->clear();
        for (auto* spinBox : m_editor->findChildren<QDoubleSpinBox*>())
            spinBox->clear();
        return;
    }


    if (materialItem->hasRefractiveIndex()) {
        m_deltaEdit->setText(QString::number(materialItem->delta().value(), 'g'));
        m_betaEdit->setText(QString::number(materialItem->beta().value(), 'g'));
    } else {
        m_realEdit->setText(QString::number(materialItem->sldRe().value(), 'g'));
        m_imaginaryEdit->setText(QString::number(materialItem->sldIm().value(), 'g'));
    }

    m_xSpinBox->setValue(materialItem->magnetization().r3().x());
    m_ySpinBox->setValue(materialItem->magnetization().r3().y());
    m_zSpinBox->setValue(materialItem->magnetization().r3().z());
}

void MaterialEditorDialog::onChangeCurrent()
{
    m_removeMaterialAction->setEnabled(currentIndex().isValid());
    m_cloneMaterialAction->setEnabled(currentIndex().isValid());

    onSelectMaterial();
    onSelectColor();
}

MaterialItem* MaterialEditorDialog::currentMaterialItem()
{
    return currentIndex().isValid() ? m_model->materialItemFromIndex(currentIndex()) : nullptr;
}

void MaterialEditorDialog::setCurrentMaterial(const MaterialItem* m)
{
    m_treeView->setCurrentIndex(m_model->indexFromMaterial(m));
    onSelectMaterial();
}

QModelIndex MaterialEditorDialog::currentIndex() const
{
    return m_treeView->currentIndex();
}

QStringList MaterialEditorDialog::identifiersOfUsedMaterials() const
{
    QStringList result;
    for (auto* p : m_sample->itemsWithMaterial())
        result << p->materialIdentifier();
    return result;
}


//! Static caller

QString GUI::chooseMaterial(SampleItem* sample, const QString& identifierOfPreviousMaterial)
{
    MaterialEditorDialog dialog(sample, identifierOfPreviousMaterial);
    if (dialog.exec() == QDialog::Accepted)
        if (MaterialItem* material = dialog.currentMaterialItem())
            return material->identifier();

    return {};
}
