//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Numeric/DoubleSpinBox.cpp
//! @brief     Implements class DoubleSpinBox.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Numeric/DoubleSpinBox.h"
#include "Base/Math/Numeric.h"
#include "Base/Util/Assert.h"
#include "Base/Util/StringUtil.h"
#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/View/Numeric/NumberUtil.h"
#include <QRegularExpression>
#include <QWheelEvent>
#include <iostream>

namespace {

static constexpr double step0 = 0.1;

QString toString(double val, int decimal_points = 6)
{
    if (val == 0)
        return "0";

    QString result = (std::abs(val) >= 10000 || std::abs(val) < 0.1)
                         ? QString::number(val, 'e', decimal_points)
                         : QString::number(val, 'f', decimal_points);

    // suppress ".0" in mantissa; normalize exponent
    return result.replace(QRegularExpression("(\\.?0+)?((e)([\\+]?)([-]?)(0*)([1-9].*))?$"),
                          "\\3\\5\\7");
}

} // namespace


DoubleSpinBox::DoubleSpinBox(DoubleProperty& d)
    : m_valueProperty(d)
    , m_step(::step0)
{
    setFocusPolicy(Qt::StrongFocus);
    GUI::View::NumberUtil::configSpinBox(this, d.decimals(), d.limits());
    setToolTip(d.tooltip());
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    QSignalBlocker b(this);
    setValue(m_valueProperty.value());

    QObject::connect(this, &QAbstractSpinBox::editingFinished, [this]() {
        setValue(fromDisplay());
        m_old_dir = 0;
        m_step = ::step0;
    });
    // setSingleStep(m_valueProperty.step());
}

QAbstractSpinBox::StepEnabled DoubleSpinBox::stepEnabled() const
{
    return StepUpEnabled | StepDownEnabled;
}

double DoubleSpinBox::fromDisplay() const
{
    double result;
    if (Base::String::to_double(lineEdit()->text().toStdString(), &result))
        return result;
    // invalid input => restore last valid value
    return m_valueProperty.value();
}

void DoubleSpinBox::setValue(double val)
{
    val = m_valueProperty.limits().clamp(val);
    lineEdit()->setText(toString(val));
    m_valueProperty.setValue(fromDisplay());
    if (val != m_valueProperty.value())
        emit valueChanged(m_valueProperty.value());
}

void DoubleSpinBox::wheelEvent(QWheelEvent* event)
{
    if (hasFocus())
        QAbstractSpinBox::wheelEvent(event);
    else
        event->ignore();
}

void DoubleSpinBox::stepBy(int steps)
{
    const int new_dir = steps > 0 ? +1 : -1;
    if (m_old_dir == new_dir)
        m_step = std::min(m_step * (std::abs(steps) == 1 ? 1.2 : 2.) * (1 + m_step), 9.);
    else if (m_old_dir == -new_dir)
        m_step = std::abs(steps) == 1 ? std::max(m_step / 9, 1e-6) : m_step;

    const double fac = (steps > 0) ^ (m_valueProperty.value() < 0) ? 1 + m_step : 1 / (1 + m_step);
    const int decimals = std::min(6, std::max(2, (int)(2 - std::log10(m_step))));

    // std::cout << "DEBUG steps=" << steps << " m_s=" << m_step << " fac=" << fac
    //	      << " dec=" << decimals << " old=" << m_valueProperty.value()
    //	      << " new=" << m_valueProperty.value() * fac
    //	      << " rnd=" << Numeric::round_decimal(m_valueProperty.value() * fac, decimals)
    //	      << std::endl;
    setValue(Numeric::round_decimal(m_valueProperty.value() * fac, decimals));

    m_old_dir = new_dir;
}

void DoubleSpinBox::updateValue()
{
    QSignalBlocker _(this);
    setValue(m_valueProperty.value());
}
